/* encode.c
 * Handlers for encoding a wav stream into a kexis stream.
 * Copyright (C) 2000 Wayde Milas (wmilas@rarcoa.com)
 *
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
  
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include "types.h"
#include "encode.h"
#include "header.h"
#include "kexis.h"
#include "predictor.h"
#include "rice.h"
#include "bits.h"

void compress(WAVHEADER *wavHead, OPTIONSTRUCT *options)
{
  int length;
	unsigned long pcmLength=0;
  KEXISBLOCKSTRUCT kexisBlock;
	PCMBLOCKSTRUCT pcmBlock;
	time_t	elapsedTime;
	PREDICTORTABLE *pred;

  options->inFileStream = fopen(options->inFileName, "r");
  if(options->inFileStream == NULL) {
    sprintf(options->errorString, "Can't open file: %s!",
			options->inFileName);
    handle_error(options);
  }

  // Suck in the WaveHeader
  parse_wave_header(wavHead, options);
  pcmLength = wavHead->data_ckSize;

	//Save the original size right now as a quick hack. Remove this later.
	pcmBlock.pcmStreamLength = wavHead->data_ckSize;

  // Since the Wav Header data length is in bytes, we need to convert it to
  // "units". Each "unit" is 2 bytes, or a short int. so we devide by 2.
  pcmLength = pcmLength / 2;

  // Here we malloc 200,000 units, or 400,000 bytes.
  pcmBlock.data = malloc(sizeof(short)*UNITS);
	kexisBlock.data = malloc(sizeof(short)*UNITS);

	// Clear the Predictor Table
	zero_predictor_table(&kexisBlock.predictor);
	kexisBlock.bitCounter=0;
	kexisBlock.bitBucket=0;
	kexisBlock.dataPosition=0;
	kexisBlock.sumTotalSize=0;
	kexisBlock.debugBlockCount=0;
	pred = &kexisBlock.predictor;

  while(pcmLength > 0) {

    if(options->progress) {
      printf("Progress:%6.2f%%   ",
        100.0-(((float)pcmLength*2.0)/(float)wavHead->data_ckSize)*100);
			time(&elapsedTime);
			printf("Time: %ld:%.2ld   ", (elapsedTime - options->time)/60,
				(elapsedTime - options->time)%60);
			if(options->encoderVersion == RICE_SINGLE_K)
				printf("K:%3d   Pd(mid,ave):%6d,%6d\r", pred->mid_K[0],
					pred->diffMid, pred->diffAve);
			else if(options->encoderVersion == RICE_DUAL_K) {
				printf("K(m,a):%3d,%3d   ", pred->mid_K[0], pred->ave_K[0]);
				printf("Pd(m,a):%6d,%6d\r", pred->diffMid, pred->diffAve);
			}
			else if(options->encoderVersion == RICE_DUAL_K_BLOCK) {
				printf("K(m,a):%3d,%3d   ",
					block_ave_K(pred->mid_K, options->kHistorySize),
					block_ave_K(pred->ave_K, options->kHistorySize));
        printf("Pd(m,a):%6d,%6d\r", pred->diffMid, pred->diffAve);
			}
      fflush(stdout);
    }

    if(pcmLength < UNITS)
      pcmBlock.size = pcmLength;
    else
      pcmBlock.size = UNITS;

    length = fread(pcmBlock.data, sizeof(short), pcmBlock.size,
			options->inFileStream);
    if(length < pcmBlock.size) {
      sprintf(options->errorString, "Error: read %d units out of %ld units.\n Error Reading PCM stream. Short Block count. Your WAV file is\n corrupt and is not as long as it should be.", length, pcmBlock.size);
      handle_error(options);
    }

    encode_pcmblock(&pcmBlock, &kexisBlock, options);

    pcmLength = pcmLength - pcmBlock.size;
  }

	cleanup_bits(&kexisBlock, options, &pcmBlock);

	handle_verbose(options, &kexisBlock, &pcmBlock);
}

void encode_pcmblock(PCMBLOCKSTRUCT *pcmBlock, KEXISBLOCKSTRUCT *kexisBlock,
	OPTIONSTRUCT *options)
{
  long   left, right, mid, ave;
  int     loop=0;

  while(loop < pcmBlock->size) {
    left = pcmBlock->data[loop];
    right = pcmBlock->data[loop+1];

    // X = average of the 2 channels
    // Y = distance from mid to right channel.
    // Thats all we need to extract L and R later, plus hopefully
    // Y is now a smaller number.
    mid = left - right;
    ave = right + (mid/2);

		// Now use our predictor to tranform X and Y into differences.
		difference_calc(kexisBlock, options->predictorVersion, mid, ave, COMPRESS);

		if(PRED_DEBUG == 1 && kexisBlock->debugBlockCount <100000) {
			printf("%ld:  %d  %d  %ld  %ld  %ld  %ld\n", kexisBlock->debugBlockCount,
				kexisBlock->predictor.diffMid, kexisBlock->predictor.diffAve,
				mid, ave, left, right);
		}	
		// If we are gathering statistics, gather them :P
		if(options->verbose)
			collect_prediction_stats(kexisBlock, kexisBlock->predictor.diffMid,
				kexisBlock->predictor.diffAve);

		// now we need to do our bit encoding
		rice_encode(kexisBlock, kexisBlock->predictor.diffMid, options,
			pcmBlock, MID);
		rice_encode(kexisBlock, kexisBlock->predictor.diffAve, options,
		  pcmBlock, AVE);

    loop = loop +2;
  }
}

void write_kexisblock(OPTIONSTRUCT *options, KEXISBLOCKSTRUCT *kexisBlock,
	PCMBLOCKSTRUCT *pcmBlock)
{
  int length;
	KEXISHEADER kexisHeader;

	// Check and see if its the beginning of the stream.
  if(options->outFileStream == NULL) {
		// Make the filename
    options->outFileName = malloc(strlen(options->inFileName)+1);
    strncpy(options->outFileName, options->inFileName,
      strlen(options->inFileName)-3);
    strcat(options->outFileName, "kxs\0");

		// Open the file
    options->outFileStream = fopen(options->outFileName, "w");
    if(options->outFileStream == NULL) {
      sprintf(options->errorString, "Can't open file: %s!",
				options->outFileName);
			handle_error(options);
		}

		// Create the header
		create_kexis_header(&kexisHeader, options, pcmBlock->pcmStreamLength);
		length = fwrite(&kexisHeader, sizeof(kexisHeader), 1,
			options->outFileStream);
		if(length != 1) {
			sprintf(options->errorString, "Error writing Header to %s!",
				options->outFileName);
			handle_error(options);
    }
  }

  length = fwrite(kexisBlock->data, sizeof(short), kexisBlock->dataPosition,
    options->outFileStream);

	if(options->progress)
		kexisBlock->sumTotalSize += length;
	
  if(length != kexisBlock->dataPosition) {
		sprintf(options->errorString, "Error writing to %s!",
			options->outFileName);
		handle_error(options);
	}
}
