/* kexis.c
 * Kexis is a lossless wav (PCM) sound compressor.
 * Copyright (C) 2000 Wayde Milas (wmilas@rarcoa.com)
 *
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <unistd.h>
#include "types.h"
#include "kexis.h"
#include "encode.h"
#include "decode.h"

void useage(void)
{
	printf("Useage:\n");
	printf("kexis [MODE] [OPTION]... FILE\n\n");
	printf("Mode:\n");
	printf(" -c   Compress a WAV file. This is the default if no mode is given.\n");
	printf("      A WAV file is the input and a KEX file is the output.\n");
	printf(" -x   Extract a KEX file to a WAV.\n");
	printf(" -s   Extract a KEX file to stdout.\n\n");
	printf("Options:\n");
	printf(" -p   Display a progress bar and ending statistics during\n");
	printf("      compression/extraction.\n");
	printf("      Default is to display NO data whatsoever.\n");
	printf(" -h   Display full WAV header if in Compress Mode.\n");
	printf("      Display full KEXIS header if in Uncompress (Extract) Mode.\n");
	printf(" -v   Display Verbose performance/debugging data during ");
	printf("procesing. Enabling\n");
	printf("      this also assumes and enables -p.\n\n");
	printf("Predictors when encoding:\n");
	printf(" -pr0 Use Dual Zero order predictor.\n");
	printf(" -pr1 Use Dual First order predictor.\n");
	printf(" -pr2 Use Dual Second order predictor. Default.\n");
	printf(" -pr3 Use Dual Third order predictor.\n");
	printf(" -pr4 Use Dual predictive Predictors. The predictor uses the best\n");
	printf("      order predictor based on the LAST values.\n\n");
	printf("Encoders used to compress:\n");
	printf(" -en1 Single K Pseudo Rice encoding.\n");
	printf(" -en2 Dual K Pseudo Rice encoding.\n");
	printf(" -en3 Dual K Pseudo Rice encoding with Variable History. Default.\n");
	printf("      The Variable history size is setable via -kh\n");
	printf(" -kh # Where # is the K history size you want to use. Default 32.\n");
	printf("      There is no \"right\" number. Optimum values differ per\n");
	printf("      individual file.\n");
	printf("\n");
	return;
}

void display_copyright(void)
{
	printf("Kexis version %d.%d.%d, Copyright (C) 2000 Wayde Milas ",
		VERSION_MAJOR, VERSION_MINOR, VERSION_SUB_MINOR);
	printf("(wmilas@rarcoa.com)\n");
	printf("Kexis comes with ABSOLUTELY NO WARRANTY; for details see included ");
	printf("GPL license\n");
	printf("This is free software, and you are welcome to redistribute it ");
	printf("under certain\nconditions; for details see included GPL license\n\n");
}

int parse_arg(OPTIONSTRUCT *options, int argc, char *argv[])
{
	int loop=1;
	char *ext=NULL;
	
	if (argc == 1) {
		display_copyright();
		useage();
		return 1;
	}

	// The default Mode
	options->inFileStream=NULL;
	options->outFileStream=NULL;
	options->inFileName=NULL;
	options->outFileName=NULL;
	options->mode = COMPRESS;
	options->progress = 0;
	options->displayHeader = 0;
	options->dec_stdout = 0;
	options->predictorVersion = STANDARD_PREDICTOR;
	options->encoderVersion = STANDARD_ENCODER;
	options->kHistorySize = DEFAULT_KHISTORY;

	while(loop < argc-1) {

		if(strcmp("-c", argv[loop]) == 0)
			options->mode = COMPRESS;
		else if(strcmp("-x", argv[loop]) == 0)
			options->mode = DECOMPRESS;
		else if(strcmp("-s", argv[loop]) == 0) {
			options->mode = DECOMPRESS; 
			options->dec_stdout = 1;
		}
		else if(strcmp("-pr0", argv[loop]) == 0)
		  options->predictorVersion = ZEROO_PREDICTOR;
		else if(strcmp("-pr1", argv[loop]) == 0)
			options->predictorVersion = FIRSTO_PREDICTOR;
		else if(strcmp("-pr2", argv[loop]) == 0)
			options->predictorVersion = SECONDO_PREDICTOR;
		else if(strcmp("-pr3", argv[loop]) == 0)
		  options->predictorVersion = THIRDO_PREDICTOR;
		else if(strcmp("-pr4", argv[loop]) == 0)
		  options->predictorVersion = L_HISTORY_PREDICTOR;
		else if(strcmp("-en1", argv[loop]) == 0)
			options->encoderVersion = RICE_SINGLE_K;
		else if(strcmp("-en2", argv[loop]) == 0)
			options->encoderVersion = RICE_DUAL_K;
		else if(strcmp("-en3", argv[loop]) == 0)
			options->encoderVersion = RICE_DUAL_K_BLOCK;
		else if(strcmp("-kh", argv[loop]) == 0) {
			loop++;
			sscanf(argv[loop], "%d", &options->kHistorySize);
			if(options->kHistorySize < 1 || options->kHistorySize > MAX_KHISTORY){
				display_copyright();
				useage();
				return -1;
			}
		}
		else if(strcmp("-p", argv[loop]) == 0)
			options->progress = 1;
		else if(strcmp("-h", argv[loop]) == 0)
      options->displayHeader = 1;
		else if(strcmp("-v", argv[loop]) == 0) {
			options->progress = 1;
			options->verbose = 1;
		}
		else {
			display_copyright();
			useage();
			return -1;
		}
		
		loop++;
	}

	if(strncmp("-", argv[loop], 1) ==0) {
		display_copyright();
		useage();
		return -1;
	}

	// Check and see if we are in compress mode that the file ends in .wav
	// and .kex is decompress mode
	ext = strrchr(argv[loop], '.');
	if(ext != NULL) {
		if(options->mode == COMPRESS) {
			if(strcmp(ext, ".wav") != 0) {
				printf("Kexis can only compress files that end in a .wav!\n");
				return -1;
			}
		}
		else {
			if(strcmp(ext, ".kxs") != 0) {
				printf("Kexis can only uncompress files that end in a .kxs!\n");
				return -1;
			}
		}
	}
	else {
		if(options->mode == COMPRESS)
			printf("Kexis can only compress files that end in a .wav!\n");
		else
			printf("Kexis can only uncompress files that end in a .kxs!\n");
		return -1;
	}

	options->inFileName = malloc(strlen(argv[loop])+1);
	strcpy(options->inFileName, argv[loop]);

	return 0;
}

int main(int argc, char *argv[])
{
	WAVHEADER 				wavHead;
	KEXISHEADER				kexisHead;
	OPTIONSTRUCT			options;
	time_t						elapsedTime;

	time(&options.time);
	if(parse_arg(&options, argc, argv))
		return 0;

	if((options.progress || options.displayHeader) && options.dec_stdout==0)
		display_copyright();

	if(options.mode == COMPRESS)
		compress(&wavHead, &options);
	
	if(options.mode == DECOMPRESS)
		decompress(&kexisHead, &options);

	if(options.progress && options.dec_stdout==0) {
		time(&elapsedTime);
		printf("\nTotal elapsed time: %ld:%.2ld \n", 
			(elapsedTime - options.time)/60,
			(elapsedTime - options.time)%60);
	}

	exit_nicely(&options);
	exit(0);
}

void handle_verbose(OPTIONSTRUCT *options, KEXISBLOCKSTRUCT *kexisBlock,
	PCMBLOCKSTRUCT *pcmBlock)
{
	struct rusage usage;

	if(options->progress && options->mode==COMPRESS && options->dec_stdout==0)
		printf("\nCompression:%6.2f%% (of original file size)",
			(float) 100.0 * 
			((float)kexisBlock->sumTotalSize/((float)pcmBlock->pcmStreamLength/2)));
	
	if(options->verbose && options->dec_stdout==0) {
		getrusage(RUSAGE_SELF,&usage);
		printf("\nTotal User Time  : %ld seconds, %ld ms.\n",
	  	usage.ru_utime.tv_sec, usage.ru_utime.tv_usec/1000);
		printf("Total System Time: %ld seconds, %ld ms.\n",
			usage.ru_stime.tv_sec, usage.ru_stime.tv_usec/1000);
		printf("Low Pd mid: %d\tHigh Pd mid: %d\tAverage Pd mid: %9.3f\n",
			kexisBlock->predictor.lowDiffMid,
			kexisBlock->predictor.highDiffMid,
			kexisBlock->predictor.sumDiffMid/kexisBlock->predictor.countDiffMid);
		printf("Low Pd ave: %d\tHigh Pd ave: %d\tAverage Pd ave: %9.3f\n",
		  kexisBlock->predictor.lowDiffAve,
			kexisBlock->predictor.highDiffAve,
			kexisBlock->predictor.sumDiffAve/
			kexisBlock->predictor.countDiffAve);
		if(options->encoderVersion == RICE_SINGLE_K)
			printf("Low K     : %d\t\tHigh K     : %d\t\tAverage K     : %9.3f",
				kexisBlock->predictor.midLowK, kexisBlock->predictor.midHighK,
				kexisBlock->predictor.midSumK/kexisBlock->predictor.midCountK);
		else if(options->encoderVersion == RICE_DUAL_K ||
			options->encoderVersion == RICE_DUAL_K_BLOCK) {
			printf("Low K mid : %d\t\tHigh K mid : %d\t\tAverage K mid : %9.3f\n",
				kexisBlock->predictor.midLowK, kexisBlock->predictor.midHighK,
				kexisBlock->predictor.midSumK/kexisBlock->predictor.midCountK);
			printf("Low K ave : %d\t\tHigh K ave : %d\t\tAverage K ave : %9.3f",
			  kexisBlock->predictor.aveLowK, kexisBlock->predictor.aveHighK,
				kexisBlock->predictor.aveSumK/kexisBlock->predictor.aveCountK);
	 	}
	}
}

void handle_error(OPTIONSTRUCT *options)
{
	printf("\n%s\n", options->errorString);
	printf("Was not able to finish. Exiting...\n");
	exit_nicely(options);
}

void exit_nicely(OPTIONSTRUCT *options)
{
	if(options->inFileStream != NULL)
		fclose(options->inFileStream);
	if(options->outFileStream != NULL)
	  fclose(options->outFileStream);
	
	printf("\n");
	exit(0);
}
